import torch
import torch.nn as nn
import numpy as np
from model import FFModel
from template import G2Model

class Predictor():
    """ Wrapper class to convert tensors to numpy and interface with G2Model, e.g. to call
        G2Model.plot_boundary()    
    """
    def __init__(self, name, model, device):
        self.name = name
        self.model = model
        self.device = device

    def score(self, x_i):
        x = torch.tensor(x_i, dtype=torch.float, device=self.device)
        scores = self.model.forward(x)
        return scores.detach().numpy()
    
    def classify(self, x_i):
        scores = self.score(x_i)
        return np.sign(scores)


def main():
    # general preparations
    smpl_size = 200
    vsmpl_size = 10000
    hdim = 500
    lrate = 1.0e-1
    name = "model_{}".format(hdim) 

    if torch.cuda.is_available():
        device = torch.device('cuda')
    else:
        device = torch.device('cpu')
    
    # get training/validation data
    # transform them to PyTorch tensors
    gmx = G2Model()
    x_i, t_i = gmx.generate_sample(smpl_size)
    x = torch.tensor(x_i, dtype=torch.float).to(device)
    t = torch.tensor(t_i, dtype=torch.int).to(device)
    xv_i, tv_i = gmx.generate_sample(vsmpl_size)
    xv = torch.tensor(xv_i, dtype=torch.float).to(device)
    tv = torch.tensor(tv_i, dtype=torch.int).to(device)

    # model
    model = FFModel(hdim, lrate, device)

    # learn
    niterations = 1000
    count = 0
    log_period = 10
    print('# Gradient checks', flush=True)
    model.check_gradient(x, t, "w1")
    model.check_gradient(x, t, "b1")
    model.check_gradient(x, t, "w2")
    model.check_gradient(x, t, "b2")
    print('# Starting', flush=True)
    while True:
        # apply gradient step
        ll = model.grad_step(x, t)
        if (count % log_period == log_period-1) or (count == niterations-1):
            with torch.no_grad():
                # get acccuracy on the validation data
                scores = model.forward(xv)
                cls = scores.sign().int()
                vacc = (cls == tv).float().mean()
            strtoprint = 'epoch: ' + str(count)
            strtoprint += ' loss: {:.4}'.format(ll)
            strtoprint += ' vacc: {:.4}'.format(vacc)
            print(strtoprint, flush=True)
        count += 1
        if count == niterations:
            break
    # visualise the learned predictor
    pred = Predictor(name, model, device)
    gmx.plot_boundary((x_i, t_i), pred)


if __name__ == '__main__':
    main()